@@ -14,7 +14,6 @@ from rasa.shared.core.constants import (
     ACTION_EXTRACT_SLOTS,
     ACTION_LISTEN_NAME,
     REQUESTED_SLOT,
-    LOOP_INTERRUPTED,
 )
 from rasa.shared.constants import UTTER_PREFIX
 from rasa.shared.core.events import (
@@ -541,7 +540,7 @@ class FormAction(LoopAction):
 
         needs_validation = (
             tracker.latest_action_name == ACTION_LISTEN_NAME
-            and not tracker.active_loop.get(LOOP_INTERRUPTED, False)
+            and not tracker.is_active_loop_interrupted
         )
 
         if needs_validation:
